"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const servie_1 = require("servie");
const error_1 = require("./error");
const universal_1 = require("servie/dist/body/universal");
/**
 * Default header handling.
 */
function normalizeRequest(options = {}) {
    return function (req, next) {
        // Block requests when already aborted.
        if (req.aborted)
            return Promise.reject(new error_1.PopsicleError('Request aborted', 'EABORT', req));
        // Remove headers that should not be created by the user.
        req.headers.delete('Host');
        // If we have no accept header set already, default to accepting
        // everything. This is needed because otherwise Firefox defaults to
        // an accept header of `html/xml`.
        if (!req.headers.get('Accept')) {
            req.headers.set('Accept', '*/*');
        }
        // Request a preference to upgrade to HTTPS.
        if (options.upgradeInsecureRequests !== false && req.Url.protocol === 'http:') {
            req.headers.set('Upgrade-Insecure-Requests', '1');
        }
        return next();
    };
}
exports.normalizeRequest = normalizeRequest;
/**
 * Redirection types to handle.
 */
var REDIRECT_TYPE;
(function (REDIRECT_TYPE) {
    REDIRECT_TYPE[REDIRECT_TYPE["FOLLOW_WITH_GET"] = 0] = "FOLLOW_WITH_GET";
    REDIRECT_TYPE[REDIRECT_TYPE["FOLLOW_WITH_CONFIRMATION"] = 1] = "FOLLOW_WITH_CONFIRMATION";
})(REDIRECT_TYPE || (REDIRECT_TYPE = {}));
/**
 * Possible redirection status codes.
 */
const REDIRECT_STATUS = {
    '301': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '302': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '303': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '307': REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION,
    '308': REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION
};
/**
 * Middleware function for following HTTP redirects.
 */
function followRedirects(options = {}) {
    return function (initialRequest, next) {
        return __awaiter(this, void 0, void 0, function* () {
            let req = initialRequest.clone();
            let redirectCount = 0;
            const maxRedirects = typeof options.maxRedirects === 'number' ? options.maxRedirects : 5;
            const confirmRedirect = options.confirmRedirect || (() => false);
            while (redirectCount++ < maxRedirects) {
                const res = yield next(req);
                const redirect = REDIRECT_STATUS[res.statusCode];
                if (redirect === undefined || !res.headers.has('Location'))
                    return res;
                const newUrl = url_1.resolve(req.url, res.headers.get('Location')); // tslint:disable-line
                // Ignore the result of the response on redirect.
                req.abort();
                req.events.emit('redirect', newUrl);
                if (redirect === REDIRECT_TYPE.FOLLOW_WITH_GET) {
                    req = initialRequest.clone();
                    req.headers.set('Content-Length', '0');
                    req.url = newUrl;
                    req.method = req.method.toUpperCase() === 'HEAD' ? 'HEAD' : 'GET';
                    req.body = universal_1.createBody(undefined);
                    req.trailer = Promise.resolve(servie_1.createHeaders());
                    continue;
                }
                if (redirect === REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION) {
                    const method = req.method.toUpperCase();
                    // Following HTTP spec by automatically redirecting with GET/HEAD.
                    if (method === 'GET' || method === 'HEAD') {
                        req = initialRequest.clone();
                        req.url = newUrl;
                        continue;
                    }
                    // Allow the user to confirm redirect according to HTTP spec.
                    if (confirmRedirect(req, res)) {
                        req = initialRequest.clone();
                        req.url = newUrl;
                        continue;
                    }
                }
                return res;
            }
            throw new error_1.PopsicleError(`Maximum redirects exceeded: ${maxRedirects}`, 'EMAXREDIRECTS', req);
        });
    };
}
exports.followRedirects = followRedirects;
//# sourceMappingURL=common.js.map