/// <reference types="node" />
import { CookieJar } from 'tough-cookie';
import { Middleware } from 'throwback';
import { Agent } from 'http';
import { SecureContext } from 'tls';
import { Request, Response, ResponseOptions } from 'servie';
import { Body } from 'servie/dist/body/node';
import { FollowRedirectsOptions, NormalizeRequestOptions } from '../common';
/**
 * Extend response with URL.
 */
export interface HttpResponseOptions extends ResponseOptions {
    url: string;
    body: Body;
    httpVersion: string;
}
/**
 * HTTP responses implement a node.js body.
 */
export declare class HttpResponse extends Response implements HttpResponseOptions {
    url: string;
    body: Body;
    httpVersion: string;
    constructor(options: HttpResponseOptions);
}
export declare class Http2Response extends HttpResponse {
}
/**
 * Cookie middleware configuration.
 */
export interface CookieOptions {
    jar: CookieJar;
}
/**
 * Read cookies from the cookie jar.
 */
export declare function getCookies<T extends Request, U extends Response>(options: CookieOptions): Middleware<T, U>;
/**
 * Save response cookies into the cookie jar.
 */
export declare function saveCookies<T extends Request, U extends Response>(options: CookieOptions): Middleware<T, U>;
/**
 * Automatically support decoding zipped responses.
 */
export declare function autoUnzip<T extends Request, U extends HttpResponse>(): Middleware<T, U>;
/**
 * Override the default user-agent header.
 */
export interface NormalizeUserAgentOptions {
    userAgent?: string;
}
/**
 * Set default user-agent in node.js.
 */
export declare function normalizeUserAgent<T extends Request, U extends Response>(options?: NormalizeUserAgentOptions): Middleware<T, U>;
export declare enum NegotiateHttpVersion {
    HTTP1_ONLY = 0,
    HTTP2_FOR_HTTPS = 1,
    HTTP2_ONLY = 2,
}
/**
 * Node.js HTTP request options.
 */
export interface SendOptions {
    rejectUnauthorized?: boolean;
    ca?: string | Buffer | Array<string | Buffer>;
    cert?: string | Buffer;
    key?: string | Buffer;
    agent?: Agent;
    secureContext?: SecureContext;
    secureProtocol?: string;
    negotiateHttpVersion?: NegotiateHttpVersion;
}
/**
 * Function to execute HTTP request.
 */
export declare function send(options: SendOptions): (req: Request) => Promise<HttpResponse>;
/**
 * Node.js HTTP transport configuration.
 */
export interface TransportOptions extends SendOptions, FollowRedirectsOptions, SendOptions, NormalizeRequestOptions, NormalizeUserAgentOptions {
    jar?: CookieJar | false;
    unzip?: false;
    follow?: false;
}
/**
 * Create a request transport using node.js `http` libraries.
 */
export declare function transport(options?: TransportOptions): (req: Request) => Promise<HttpResponse>;
