"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const pump = require("pump");
const tough_cookie_1 = require("tough-cookie");
const throwback_1 = require("throwback");
const http_1 = require("http");
const https_1 = require("https");
const net_1 = require("net");
const tls_1 = require("tls");
const http2_1 = require("http2");
const stream_1 = require("stream");
const zlib_1 = require("zlib");
const servie_1 = require("servie");
const node_1 = require("servie/dist/body/node");
const error_1 = require("../error");
const common_1 = require("../common");
/**
 * HTTP responses implement a node.js body.
 */
class HttpResponse extends servie_1.Response {
    constructor(options) {
        super(options);
        this.url = options.url;
        this.body = options.body;
        this.httpVersion = options.httpVersion;
    }
}
exports.HttpResponse = HttpResponse;
class Http2Response extends HttpResponse {
}
exports.Http2Response = Http2Response;
/**
 * Read cookies from the cookie jar.
 */
function getCookies(options) {
    return function (req, next) {
        const prevCookies = req.headers.getAll('Cookie').join('; ');
        return new Promise((resolve, reject) => {
            options.jar.getCookieString(req.url, (err, cookies) => {
                if (err)
                    return reject(err);
                if (cookies) {
                    req.headers.set('Cookie', prevCookies ? `${prevCookies}; ${cookies}` : cookies);
                }
                return resolve(next());
            });
        });
    };
}
exports.getCookies = getCookies;
/**
 * Save response cookies into the cookie jar.
 */
function saveCookies(options) {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield next();
            const cookies = res.headers.getAll('set-cookie');
            if (cookies.length) {
                yield Promise.all(cookies.map(function (cookie) {
                    return new Promise(function (resolve, reject) {
                        options.jar.setCookie(cookie, req.url, { ignoreError: true }, (err) => err ? reject(err) : resolve());
                    });
                }));
            }
            return res;
        });
    };
}
exports.saveCookies = saveCookies;
/**
 * Automatically support decoding zipped responses.
 */
function autoUnzip() {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!req.headers.has('Accept-Encoding')) {
                req.headers.set('Accept-Encoding', 'gzip,deflate');
            }
            const res = yield next();
            const enc = res.headers.get('Content-Encoding');
            // Unzip body automatically when response is encoded.
            if (enc === 'deflate' || enc === 'gzip') {
                res.body = node_1.createBody(pump(res.body.stream(), zlib_1.createUnzip()), { headers: [] });
            }
            return res;
        });
    };
}
exports.autoUnzip = autoUnzip;
/**
 * Set default user-agent in node.js.
 */
function normalizeUserAgent(options = {}) {
    const userAgent = options.userAgent || 'Popsicle (https://github.com/serviejs/popsicle)';
    return function (req, next) {
        if (!req.headers.has('User-Agent'))
            req.headers.set('User-Agent', userAgent);
        return next();
    };
}
exports.normalizeUserAgent = normalizeUserAgent;
var NegotiateHttpVersion;
(function (NegotiateHttpVersion) {
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP1_ONLY"] = 0] = "HTTP1_ONLY";
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP2_FOR_HTTPS"] = 1] = "HTTP2_FOR_HTTPS";
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP2_ONLY"] = 2] = "HTTP2_ONLY";
})(NegotiateHttpVersion = exports.NegotiateHttpVersion || (exports.NegotiateHttpVersion = {}));
/**
 * Write Servie body object to node.js stream.
 */
function sendBody(body, stream, onError) {
    // Pipe the body to the stream.
    if (body instanceof node_1.Body) {
        if (body.buffered) {
            body.buffer().then(x => stream.end(x), onError);
        }
        else {
            pump(body.stream(), stream);
        }
    }
    else {
        body.arrayBuffer().then(x => stream.end(Buffer.from(x)), onError);
    }
}
/**
 * Execute HTTP request.
 */
function execHttp(req, protocol, host, port, socket, agent) {
    return new Promise((resolve, reject) => {
        const { url, body, Url } = req;
        const encrypted = Url.protocol === 'https:';
        const request = encrypted ? https_1.request : http_1.request;
        const arg = {
            protocol,
            host,
            port,
            method: req.method,
            path: Url.path,
            headers: req.headers.asObject(false),
            auth: Url.auth,
            agent,
            createConnection: () => socket
        };
        const rawRequest = request(arg);
        const requestStream = new stream_1.PassThrough();
        // Trigger unavailable error when node.js errors before response.
        function onError(err) {
            return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
        }
        // Track the node.js response.
        function onResponse(rawResponse) {
            const { statusCode, statusMessage, httpVersion } = rawResponse;
            const headers = servie_1.createHeaders(rawResponse.rawHeaders);
            const body = node_1.createBody(pump(rawResponse, new stream_1.PassThrough()), { headers: {} });
            // Trailers are populated on "end".
            const trailer = new Promise(resolve => {
                rawResponse.on('end', () => resolve(servie_1.createHeaders(rawResponse.rawTrailers)));
            });
            // Replace request error listener behaviour.
            rawRequest.removeListener('error', onError);
            rawRequest.on('error', err => req.events.emit('error', err));
            const { address: localAddress, port: localPort } = rawRequest.connection.address();
            const { address: remoteAddress, port: remotePort } = rawResponse.connection.address();
            const res = new HttpResponse({ statusCode, statusMessage, headers, trailer, body, url, httpVersion });
            // Update request connection.
            req.connection = { localAddress, localPort, remoteAddress, remotePort, encrypted };
            // https://github.com/serviejs/servie#implementers
            res.started = true;
            req.events.emit('response', res);
            // Track response progress.
            rawResponse.on('data', (chunk) => res.bytesTransferred += chunk.length);
            rawResponse.on('end', () => res.finished = true);
            rawResponse.on('close', () => req.closed = true);
            return resolve(res);
        }
        rawRequest.once('error', onError);
        rawRequest.once('response', onResponse);
        // https://github.com/serviejs/servie#implementers
        req.started = true;
        req.events.on('abort', () => rawRequest.abort());
        // Track request upload progress.
        requestStream.on('data', (chunk) => req.bytesTransferred += chunk.length);
        pump(requestStream, rawRequest, () => req.finished = true);
        return sendBody(body, requestStream, reject);
    });
}
/**
 * Execute a HTTP2 connection.
 */
function execHttp2(req, protocol, host, port, socket) {
    return new Promise((resolve, reject) => {
        // HTTP2 formatted headers.
        const headers = Object.assign(req.headers.asObject(false), {
            [http2_1.constants.HTTP2_HEADER_METHOD]: req.method,
            [http2_1.constants.HTTP2_HEADER_PATH]: req.Url.path
        });
        // TODO: Fix node.js types.
        const connectOptions = {
            createConnection: () => socket
        };
        const authority = `${protocol}//${host}:${port}`;
        const client = http2_1.connect(authority, connectOptions);
        const http2Stream = client.request(headers, { endStream: false });
        const requestStream = new stream_1.PassThrough();
        // Trigger unavailable error when node.js errors before response.
        function onError(err) {
            return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
        }
        function onResponse(headers) {
            const encrypted = socket.encrypted === true;
            const { localAddress, localPort, remoteAddress, remotePort } = socket;
            // Replace request error listener behaviour with proxy.
            http2Stream.removeListener('error', onError);
            http2Stream.on('error', err => req.events.emit('error', err));
            const res = new Http2Response({
                statusCode: Number(headers[http2_1.constants.HTTP2_HEADER_STATUS]),
                url: req.url,
                httpVersion: '2.0',
                headers: servie_1.createHeaders(headers),
                body: node_1.createBody(pump(http2Stream, new stream_1.PassThrough()), { headers: {} })
            });
            // https://github.com/serviejs/servie#implementers
            res.started = true;
            req.connection = { localAddress, localPort, remoteAddress, remotePort, encrypted };
            // Track response progress.
            http2Stream.on('data', (chunk) => res.bytesTransferred += chunk.length);
            // Close HTTP2 session when request ends.
            http2Stream.on('end', () => {
                req.closed = true;
                res.finished = true;
                client.close();
            });
            return resolve(res);
        }
        client.once('error', onError);
        http2Stream.once('error', onError);
        http2Stream.once('response', onResponse);
        // https://github.com/serviejs/servie#implementers
        req.started = true;
        req.events.on('abort', () => http2Stream.destroy());
        // Track request upload progress.
        requestStream.on('data', (chunk) => req.bytesTransferred += chunk.length);
        pump(requestStream, http2Stream, () => req.finished = true);
        return sendBody(req.body, requestStream, reject);
    });
}
/**
 * Function to execute HTTP request.
 */
function send(options) {
    // Mirror common browser behaviour by default.
    const { negotiateHttpVersion = NegotiateHttpVersion.HTTP2_FOR_HTTPS } = options;
    return function (req) {
        const { hostname, protocol } = req.Url;
        const host = hostname || 'localhost';
        if (protocol === 'http:') {
            const port = Number(req.Url.port) || 80;
            const socketOptions = { host, port };
            if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                return execHttp2(req, protocol, host, port, net_1.connect(socketOptions));
            }
            return execHttp(req, protocol, host, port, net_1.connect(socketOptions), options.agent);
        }
        // Optionally negotiate HTTP2 connection.
        if (protocol === 'https:') {
            const port = Number(req.Url.port) || 443;
            const socketOptions = {
                host,
                port,
                servername: calculateServerName(host, req.headers.get('host')),
                rejectUnauthorized: options.rejectUnauthorized !== false,
                ca: options.ca,
                cert: options.cert,
                key: options.key,
                secureProtocol: options.secureProtocol,
                secureContext: options.secureContext
            };
            if (negotiateHttpVersion === NegotiateHttpVersion.HTTP1_ONLY) {
                return execHttp(req, protocol, host, port, tls_1.connect(socketOptions), options.agent);
            }
            if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                socketOptions.ALPNProtocols = ['h2']; // Only requesting HTTP2 support.
                return execHttp2(req, protocol, host, port, tls_1.connect(socketOptions));
            }
            return new Promise((resolve, reject) => {
                socketOptions.ALPNProtocols = ['h2', 'http/1.1']; // Request HTTP2 or HTTP1.
                const socket = tls_1.connect(socketOptions);
                socket.once('secureConnect', () => {
                    const alpnProtocol = socket.alpnProtocol;
                    // Successfully negotiated HTTP2 connection.
                    if (alpnProtocol === 'h2') {
                        return resolve(execHttp2(req, protocol, host, port, socket));
                    }
                    if (alpnProtocol === 'http/1.1') {
                        return resolve(execHttp(req, protocol, host, port, socket, options.agent));
                    }
                    return reject(new error_1.PopsicleError('No ALPN protocol negotiated', 'EALPNPROTOCOL', req));
                });
                socket.once('error', (err) => {
                    return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
                });
            });
        }
        return Promise.reject(new error_1.PopsicleError(`Unsupported URL protocol: ${req.Url.protocol}`, 'EPROTOCOL', req));
    };
}
exports.send = send;
/**
 * Create a request transport using node.js `http` libraries.
 */
function transport(options = {}) {
    const fns = [common_1.normalizeRequest(options), normalizeUserAgent(options)];
    const { jar = new tough_cookie_1.CookieJar(), unzip = true, follow = true } = options;
    if (unzip)
        fns.push(autoUnzip());
    if (follow)
        fns.push(common_1.followRedirects(options));
    if (jar)
        fns.push(getCookies({ jar }), saveCookies({ jar }));
    const done = send(options);
    const middlware = throwback_1.compose(fns);
    return (req) => middlware(req, done);
}
exports.transport = transport;
/**
 * Ref: https://github.com/nodejs/node/blob/5823938d156f4eb6dc718746afbf58f1150f70fb/lib/_http_agent.js#L231
 */
function calculateServerName(host, hostHeader) {
    if (!hostHeader)
        return host;
    if (hostHeader.charAt(0) === '[') {
        const index = hostHeader.indexOf(']');
        if (index === -1)
            return hostHeader;
        return hostHeader.substr(1, index - 1);
    }
    return hostHeader.split(':', 1)[0];
}
//# sourceMappingURL=http.js.map