import { Request, Response, ResponseOptions } from 'servie';
import { Body } from 'servie/dist/body/browser';
/**
 * Extend response with URL.
 */
export interface XhrResponseOptions extends ResponseOptions {
    url: string;
    body: Body;
}
/**
 * XHR responses can indicate a response URL.
 */
export declare class XhrResponse extends Response implements XhrResponseOptions {
    url: string;
    body: Body;
    constructor(options: XhrResponseOptions);
}
/**
 * Valid XHR configuration.
 */
export interface SendOptions {
    type?: XMLHttpRequestResponseType;
    withCredentials?: boolean;
    overrideMimeType?: string;
}
/**
 * Send request over `XMLHttpRequest`.
 */
export declare function send(options: SendOptions): (req: Request) => Promise<XhrResponse>;
/**
 * Combined transport options.
 */
export interface TransportOptions extends SendOptions {
}
/**
 * Create a request transport using `XMLHttpRequest`.
 */
export declare function transport(options?: TransportOptions): (req: Request) => Promise<XhrResponse>;
